@@ -32,6 +32,7 @@ int main(int argc, char *argv[])
 	KP = rq_KP(rq);
 	N = rq_N(rq);
 	Z = rq_Z(rq);
+	Al = rq_Al(rq);
 
 	/* free RaptorQ context */
 	rq_free(rq);
@@ -42,6 +43,7 @@ int main(int argc, char *argv[])
 	printf("K' %12u\tnumber of symbols, including padding symbols\n", KP);
 	printf("N  %12u\tnumber of sub-blocks\n", N);
 	printf("Z  %12u\tnumber of blocks\n", Z);
+	printf("Al %12u\talignment\n", Al);
 
 	return 0;
 }
