@@ -1,5 +1,5 @@
 /* SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only */
-/* Copyright (c) 2022 Brett Sheffield <bacs@librecast.net> */
+/* Copyright (c) 2022-2024 Brett Sheffield <bacs@librecast.net> */
 
 #include <matrix.h>
 #include <gf256.h>
@@ -10,6 +10,22 @@
 
 #define VSZ 256
 
+static void matrix_row_add_dispatch(matrix_t *dst, const int drow, const matrix_t *src, const int srow);
+void (*matrix_row_add)(matrix_t *, const int, const matrix_t *, const int) = &matrix_row_add_dispatch;
+
+static void matrix_row_mul_dispatch(matrix_t *m, const int row, const int off, const uint8_t y);
+void (*matrix_row_mul)(matrix_t *, const int, const int, const uint8_t) = &matrix_row_mul_dispatch;
+
+static void matrix_row_mul_byrow_dispatch(matrix_t *m, const int rdst, const int off, const int rsrc, const uint8_t y);
+void (*matrix_row_mul_byrow)(matrix_t *, const int, const int, const int, const uint8_t) = &matrix_row_mul_byrow_dispatch;
+
+#if !(USE_NATIVE)
+void matrix_row_add_avx2(matrix_t *dst, const int drow, const matrix_t *src, const int srow);
+void matrix_row_add_sse2(matrix_t *dst, const int drow, const matrix_t *src, const int srow);
+void matrix_row_mul_ssse3(matrix_t *m, const int row, const int off, const uint8_t y);
+void matrix_row_mul_byrow_ssse3(matrix_t *m, const int rdst, const int off, const int rsrc, const uint8_t y);
+#endif /* USE_NATIVE */
+
 uint8_t reclen[5] = {
 	0, /* NOOP */
 	sizeof(matrix_op_swap_t),
@@ -241,36 +257,66 @@ matrix_t *matrix_swap_cols(matrix_t *m, const int c1, const int c2)
 	}
 	return m;
 }
+#if USE_NATIVE
+# include "matrix_avx2.c"
+# include "matrix_ssse3.c"
+# include "matrix_sse2.c"
+#endif /* USE_NATIVE */
 
-void matrix_row_add(matrix_t *dst, const int drow, const matrix_t *src, const int srow)
+static void matrix_row_add_nosimd(matrix_t *dst, const int drow, const matrix_t *src, const int srow)
 {
 	assert(matrix_cols(dst) == matrix_cols(src));
 	uint8_t *d = matrix_ptr_row(dst, drow);
 	uint8_t *s = matrix_ptr_row(src, srow);
 	const int mcols = matrix_cols(dst);
-	int j;
-#if   defined(INTEL_AVX2)
-	const int mod_avx2 = mcols % 32;
-	const int maxv_avx2 = mcols - mod_avx2;
-	for (j = 0; j < maxv_avx2; j += 32) {
-		__m256i S = _mm256_loadu_si256((const __m256i_u *)&s[j]);
-		__m256i D = _mm256_loadu_si256((const __m256i_u *)&d[j]);
-		D = _mm256_xor_si256(D, S);
-		_mm256_storeu_si256((__m256i*)&d[j], D);
-	}
-#elif defined(INTEL_SSE3)
-	const int mod_sse3 = mcols % 16;
-	const int maxv_sse3 = mcols - mod_sse3;
-	for (j = 0; j < maxv_sse3; j += 16) {
-		__m128i S = _mm_loadu_si128((const __m128i_u *)&s[j]);
-		__m128i D = _mm_loadu_si128((const __m128i_u *)&d[j]);
-		D = _mm_xor_si128(D, S);
-		_mm_storeu_si128((__m128i*)&d[j], D);
-	}
-#else
-	j = 0;
+	for (int j = 0; j < mcols; j++) d[j] ^= s[j];
+}
+
+void matrix_row_mul_nosimd(matrix_t *m, const int row, const int off, const uint8_t y)
+{
+	uint8_t *d = matrix_ptr_row(m, row) + off;
+	const int max = m->cols - off;
+	for (int j = 0; j < max; j++) d[j] = GF256MUL(d[j], y);
+}
+
+void matrix_row_mul_byrow_nosimd(matrix_t *m, const int rdst, const int off, const int rsrc, const uint8_t y)
+{
+	assert(y);
+	uint8_t *d = matrix_ptr_row(m, rdst) + off;
+	uint8_t *s = matrix_ptr_row(m, rsrc) + off;
+	const int max = m->cols - off;
+	for (int i = 0; i < max; i++) d[i] ^= GF256MUL(s[i], y);
+}
+
+static void matrix_row_add_dispatch(matrix_t *dst, const int drow, const matrix_t *src, const int srow)
+{
+	matrix_row_add = &matrix_row_add_nosimd;
+#ifdef USE_SIMD_x86
+	const int isets = cpu_instruction_set();
+	if      (isets & AVX2) matrix_row_add = &matrix_row_add_avx2;
+	else if (isets & SSE2) matrix_row_add = &matrix_row_add_sse2;
 #endif
-	for (; j < mcols; j++) d[j] ^= s[j];
+	matrix_row_add(dst, drow, src, srow);
+}
+
+static void matrix_row_mul_dispatch(matrix_t *m, const int row, const int off, const uint8_t y)
+{
+	matrix_row_mul = &matrix_row_mul_nosimd;
+#ifdef USE_SIMD_x86
+	const int isets = cpu_instruction_set();
+	if (isets & SSSE3) matrix_row_mul = &matrix_row_mul_ssse3;
+#endif
+	matrix_row_mul(m, row, off, y);
+}
+
+static void matrix_row_mul_byrow_dispatch(matrix_t *m, const int rdst, const int off, const int rsrc, const uint8_t y)
+{
+	matrix_row_mul_byrow = &matrix_row_mul_byrow_nosimd;
+#ifdef USE_SIMD_x86
+	const int isets = cpu_instruction_set();
+	if (isets & SSSE3) matrix_row_mul_byrow = &matrix_row_mul_byrow_ssse3;
+#endif
+	matrix_row_mul_byrow(m, rdst, off, rsrc, y);
 }
 
 matrix_t matrix_add(const matrix_t *x, const matrix_t *y)
@@ -408,73 +454,6 @@ void matrix_inverse_LU(matrix_t *IA, const matrix_t *LU, const int P[])
 	}
 }
 
-#ifdef INTEL_SSE3
-inline static __m128i mul_128(__m128i A, uint8_t y)
-{
-	__m128i table1 = _mm_loadu_si128((const __m128i_u *)GF256LR[y][0]);
-	__m128i table2 = _mm_loadu_si128((const __m128i_u *)GF256LR[y][1]);
-	__m128i mask1 = _mm_set1_epi8((uint8_t)0x0f);
-	__m128i mask2 = _mm_set1_epi8((uint8_t)0xf0);
-	__m128i l, h;
-
-	l = _mm_and_si128(A, mask1);
-	l = _mm_shuffle_epi8(table1, l);
-	h = _mm_and_si128(A, mask2);
-	h = _mm_srli_epi64(h, 4);
-	h = _mm_shuffle_epi8(table2, h);
-	return _mm_xor_si128(h, l);
-}
-#endif
-
-void matrix_row_mul(matrix_t *m, const int row, const int off, const uint8_t y)
-{
-	uint8_t *d = matrix_ptr_row(m, row) + off;
-	const int max = m->cols - off;
-#ifdef INTEL_SSE3
-	const int mod = max % 16;
-	const int maxv = max - mod;
-	int j;
-
-	for (j = 0; j < maxv; j += 16) {
-		__m128i D = _mm_loadu_si128((const __m128i_u *)&d[j]);
-		D = mul_128(D, y);
-		_mm_storeu_si128((__m128i*)&d[j], D);
-	}
-#else
-	int j = 0;
-#endif
-	for (; j < max; j++) {
-		d[j] = GF256MUL(d[j], y);
-	}
-}
-
-void matrix_row_mul_byrow(matrix_t *m, const int rdst, const int off, const int rsrc, const uint8_t y)
-{
-	assert(y);
-	uint8_t *d = matrix_ptr_row(m, rdst) + off;
-	uint8_t *s = matrix_ptr_row(m, rsrc) + off;
-#ifdef INTEL_SSE3
-	const int max = m->cols - off;
-	const int mod = max % 16;
-	const int maxv = max - mod;
-	int i;
-
-	for (i = 0; i < maxv; i += 16) {
-		__m128i S = _mm_loadu_si128((const __m128i_u *)&s[i]);
-		__m128i D = _mm_loadu_si128((const __m128i_u *)&d[i]);
-		S = mul_128(S, y);
-		D = _mm_xor_si128(D, S);
-		_mm_storeu_si128((__m128i*)&d[i], D);
-	}
-#else
-	const int max = m->cols - off;
-	int i = 0;
-#endif
-	for (; i < max; i++) {
-		d[i] ^= GF256MUL(s[i], y);
-	}
-}
-
 void matrix_solve_LU(matrix_t *X, const matrix_t *Y, const matrix_t *LU, const int P[], const int Q[])
 {
 	int n = MIN(matrix_rows(LU), matrix_cols(LU));
